/*
 * mysql.c	Functions to access the MYSQL database. 
 *
 * Much of the Mysql connection and accounting code was taken from 
 * Wim Bonis's (bonis@kiss.de) accounting patch to livingston radius
 * 2.01. His patch can be found at:
 *
 *       ftp://ftp.kiss.de/pub/unix/livingston/mysql-patches.tgz
 *
 * Version:	@(#)mysql.c  1.10  25-Sep-1998  oyarzun@wilmington.net
 *
 */

#ifdef USEMYSQL 

#include	<sys/types.h>
#include	<sys/socket.h>
#include	<sys/time.h>
#include	<netinet/in.h>

#include	<stdio.h>
#include	<stdlib.h>
#include	<netdb.h>
#include	<pwd.h>
#include	<time.h>
#include	<ctype.h>
#include	<strings.h>

#include        <mysql/mysql.h>


#include	"radiusd.h"

#define MYSQL_AUTH 1
#define MYSQL_ACCT 2

MYSQL	*mysql_Reconnect(int);
int	my_mysql_query(const char *,MYSQL *,int);
static char	*make_query(char *, char *);
static void	fieldcpy(char *, char **);
char	*crypt(const char *, const char *);

MYSQL real_mysql_auth, *mysql_auth = NULL;
MYSQL real_mysql_acct, *mysql_acct = NULL;

char mysql_server[40];
char mysql_login[20];
char mysql_password[20];
char acct_db[20];
char auth_db[20];
char auth_query[MAX_AUTH_QUERY_LEN];
char mysql_auth_table[20];
char mysql_acct_table[20];
int  mysql_keepopen;
int  use_mysql_acct;
int  use_mysql_auth;


/*************************************************************************
 *
 *	Function: mysql_rad_init
 *
 *	Purpose: Reads in MySQL Config File 
 *
 *************************************************************************/

int mysql_rad_init ()
{
	FILE    *sqlfd;
        char    dummystr[64];
        char    namestr[64];
        int     line_no;
        char    buffer[256];
        char    sqlfile[256];
        char    *ptr;
       
       strcpy(mysql_server,"");
       strcpy(mysql_login,"");
       strcpy(mysql_password,"");
       strcpy(acct_db,"");
       strcpy(auth_db,"");
       strcpy(mysql_auth_table,"passwd");
       strcpy(mysql_acct_table,"usertime");
       mysql_keepopen = 0;
       use_mysql_acct = 0;
       use_mysql_auth = 0;

        sprintf(sqlfile, "%s/%s", radius_dir, "sqlserver");
        if((sqlfd = fopen(sqlfile, "r")) == (FILE *)NULL) {
                log(L_ERR,"could not read sqlserver file %s",sqlfile);
                return(-1);
        }

        line_no = 0;
        while(fgets(buffer, sizeof(buffer), sqlfd) != (char *)NULL) {
                line_no++;

                /* Skip empty space */
                if(*buffer == '#' || *buffer == '\0' || *buffer == '\n') {
                        continue;
                }

                if(strncasecmp(buffer, "server", 6) == 0) {
                        /* Read the SERVER line */
                        if(sscanf(buffer, "%s%s", dummystr, namestr) != 2) {
                               log(L_ERR,"invalid attribute on line %d of sqlserver file %s", line_no,sqlfile);
                         use_mysql_acct = 0;
                         use_mysql_auth = 0;
                       } else {
                         strcpy(mysql_server,namestr);
                       }
               }
                if(strncasecmp(buffer, "login", 5) == 0) {
                        /* Read the LOGIN line */
                        if(sscanf(buffer, "%s%s", dummystr, namestr) != 2) {
                               log(L_ERR,"invalid attribute on line %d of sqlserver file %s", line_no,sqlfile);
                       } else {
                         strcpy(mysql_login,namestr);
                       }
               }
                if(strncasecmp(buffer, "password", 8) == 0) {
                        /* Read the PASSWORD line */
                        if(sscanf(buffer, "%s%s", dummystr, namestr) != 2) {
                               log(L_ERR,"invalid attribute on line %d of sqlserver file %s", line_no,sqlfile);
                       } else {
                         strcpy(mysql_password,namestr);
                       }
               }
                if(strncasecmp(buffer, "acct_db", 7) == 0) {
                        /* Read the ACCT_DB line */
                        if(sscanf(buffer, "%s%s", dummystr, namestr) != 2) {
                               log(L_ERR,"invalid attribute on line %d of sqlserver file %s", line_no,sqlfile);
                         use_mysql_acct = 0;
                       } else {
                         strcpy(acct_db,namestr);
                       }
               }
                if(strncasecmp(buffer, "auth_db", 7) == 0) {
                        /* Read the AUTH_DB line */
                        if(sscanf(buffer, "%s%s", dummystr, namestr) != 2) {
                               log(L_ERR,"invalid attribute on line %d of sqlserver file %s", line_no,sqlfile);
                         use_mysql_auth = 0;
                       } else {
                         strcpy(auth_db,namestr);
                       }
               }
                if(strncasecmp(buffer, "auth_table", 10) == 0) {
                        /* Read the AUTH_TABLE line */
                        if(sscanf(buffer, "%s%s", dummystr, namestr) != 2) {
                               log(L_ERR,"invalid attribute on line %d of sqlserver file %s", line_no,sqlfile);
                         use_mysql_auth = 0;
                       } else {
                         strcpy(mysql_auth_table,namestr);
                       }
               }
                if(strncasecmp(buffer, "acct_table", 10) == 0) {
                        /* Read the ACCT_TABLE line */
                        if(sscanf(buffer, "%s%s", dummystr, namestr) != 2) {
                               log(L_ERR,"invalid attribute on line %d of sqlserver file %s", line_no,sqlfile);
                         use_mysql_acct = 0;
                       } else {
                         strcpy(mysql_acct_table,namestr);
                       }
               }
                if(strncasecmp(buffer, "auth_query", 10) == 0) {
                        /* Read the ACCT_QUERY line
                        if(sscanf(buffer, "%s%s", dummystr, namestr) != 2) {
                               log(L_ERR,"invalid attribute on line %d of sqlserver file %s", line_no,sqlfile);
                         use_mysql_auth = 0;
                       } else { */
                         ptr = buffer + 10;
                         fieldcpy(auth_query,&ptr);
                         /*strcpy(auth_query,namestr);
                       }*/
               }

                if(strncasecmp(buffer, "keepopen", 8) == 0) {
                        /* Read the KEEPOPEN line */
                        if(sscanf(buffer, "%s%s", dummystr, namestr) != 2) {
                               log(L_ERR,"invalid attribute on line %d of sqlserver file %s", line_no,sqlfile);
                       } else {
                         if(strncasecmp(namestr, "yes", 3) == 0) {
                           mysql_keepopen = 1;
                         } else {
                           mysql_keepopen = 0;
                         }
                       }
               }
                if(strncasecmp(buffer, "doauth", 6) == 0) {
                        /* Read the KEEPOPEN line */
                        if(sscanf(buffer, "%s%s", dummystr, namestr) != 2) {
                               log(L_ERR,"invalid attribute on line %d of sqlserver file %s", line_no,sqlfile);
                       } else {
                         if(strncasecmp(namestr, "yes", 3) == 0) {
                           use_mysql_auth = 1;
                         } else {
                           use_mysql_auth = 0;
                         }
                       }
               }
                if(strncasecmp(buffer, "doacct", 6) == 0) {
                        /* Read the KEEPOPEN line */
                        if(sscanf(buffer, "%s%s", dummystr, namestr) != 2) {
                               log(L_ERR,"invalid attribute on line %d of sqlserver file %s", line_no,sqlfile);
                       } else {
                         if(strncasecmp(namestr, "yes", 3) == 0) {
                           use_mysql_acct = 1;
                         } else {
                           use_mysql_acct = 0;
                         }
                       }
               }
       }
       fclose(sqlfd);
       
       log(L_INFO,"MYSQL_init: using: %s,%s,%s,%s,%s,%s,%d,%d,%d",
       mysql_server,
       mysql_login,
       acct_db,
       auth_db,
       mysql_auth_table,
       mysql_acct_table,
       mysql_keepopen,
       use_mysql_acct,
       use_mysql_auth); 
           
       return 0;
}


/*************************************************************************
 *
 *	Function: build_mysql_record
 *
 *	Purpose: Builds accounting record for MySQL 
 *
 *************************************************************************/

void build_mysql_record(VALUE_PAIR * pair, MYSQL_RECORD * rec)
{
       char            buffer[100];
       char           *str;

       
       if (use_mysql_acct == 0) return;
       switch (pair->attribute) {
       case PW_ACCT_SESSION_ID:
               strncpy(rec->id, pair->strvalue,8);
               break;
       case PW_USER_NAME:
               if (pair->strvalue[0] <= 'Z') { /* Strip first Upper case
                                                * Letter */
                       str = pair->strvalue + 1;
                       strcpy(rec->name, str);
               } else {
                       strcpy(rec->name, pair->strvalue);
               }
               break;
       case PW_ACCT_STATUS_TYPE:
               rec->start = pair->lvalue;
               break;
       case PW_NAS_IP_ADDRESS:
               ipaddr2str(buffer, pair->lvalue);
               strcpy(rec->server, buffer);
               /*rec->server = pair->lvalue; */;   
               /* IP addresse des Terminal Servers */
               break;
       case PW_FRAMED_IP_ADDRESS:
               ipaddr2str(buffer, pair->lvalue);
               strcpy(rec->ip, buffer);
               /* rec->ip = pair->lvalue; */ /* IP addresse des Users */
               break;
       case PW_NAS_PORT_ID:
               rec->port = pair->lvalue;
               break;
       case PW_SERVICE_TYPE:      /* Service-Type = Framed-User */
               rec->service_type = pair->lvalue;
               break;
       case PW_FRAMED_PROTOCOL:
               rec->protocol = pair->lvalue;
               break;
       case PW_LOGIN_SERVICE:
               rec->protocol = pair->lvalue + 100;
               break;
       case PW_ACCT_SESSION_TIME:
               rec->time_used = pair->lvalue;
               break;
       case PW_ACCT_AUTHENTIC:
               rec->auth = pair->lvalue;
               break;
       case PW_ACCT_OUTPUT_OCTETS:
               rec->outbytes = pair->lvalue;
               break;
       case PW_ACCT_INPUT_OCTETS:
               rec->inbytes = pair->lvalue;
               break;
               /*
                * case PW_TERMINATION: rec->term_cause = pair->lvalue;
                * break;
                */
       case PW_LOGIN_IP_HOST:     /* zb UUCP Logins per UUCP */
               ipaddr2str(buffer, pair->lvalue);
               strcpy(rec->ip, buffer);
               /* rec->ip = pair->lvalue;*/ /* IP addresse des Hosts */
               break;
       case PW_ACCT_DELAY_TIME:
               rec->delay = pair->lvalue;
               break;
       case PW_FRAMED_COMPRESSION:
               break;
       case PW_ACCT_INPUT_PACKETS:			   
               break;
       case PW_ACCT_OUTPUT_PACKETS:			   
               break;
       case PW_NAS_PORT_TYPE:			   
               rec->port_type = pair->lvalue;
               break;
       case PW_CONNECT_INFO:			   
               strcpy(rec->connect_info, pair->strvalue);
               break;
       default:
               if (!strcasecmp(pair->name, "called-station-id")) {
                       strcpy(rec->call_to, pair->strvalue);
               } else if (!strcasecmp(pair->name, "calling-station-id")) {
                       int             count = 0;
                       int             i = 0;
                       str = pair->strvalue;
                       while ( (*str != '\0') && (i < 3) ) {
                               if ((*str >= '\200') && (*str <= '\205')) {     
/* Remove Leading
 * \200\201 */
                                       count++;
                               }
                               str++;
                               i++;
                       }
                       strcpy(rec->call_from, pair->strvalue + count);
               } else if (!strcasecmp(pair->name, "Acct-Terminate-Cause")) {
                       rec->term_cause = pair->lvalue;
               } else if (!strcasecmp(pair->name, "NAS_Identifier")) {
		      /*do nothing to elminate error below show up w/ a USR TC*/
               } else {
                       log(L_ERR,"Unkown Pair: %s", pair->name);
               }
       }
}


/*************************************************************************
 *
 *	Function: mysql_Reconnect
 *
 *	Purpose: Connect to MySQL Server
 *
 *************************************************************************/

/*
 * FIXME: There is probably a better way of implenting this function to
 *        connect to two db's.
 */

MYSQL *mysql_Reconnect(int type)
{
  int i ;
  MYSQL * mysql = NULL;

  i = 0; 
  while(i<10) { // Try it 10 Times 
     if (type == MYSQL_AUTH) {
       if (!(mysql_connect(&real_mysql_auth, mysql_server,mysql_login,mysql_password
))) {
               log(L_ERR,"MYSQL: Cannot Connect to %s as %s",mysql_server
,mysql_login);
               mysql = NULL;
       } else {
               // log(L_ERR,"MYSQL Connected to %s",mysql_server);
               mysql = &real_mysql_auth;
               break;
       }
     } else {
       if (!(mysql_connect(&real_mysql_acct, mysql_server,mysql_login,mysql_password
))) {
               log(L_ERR,"MYSQL: Cannot Connect to %s as %s",mysql_server
,mysql_login);
               mysql = NULL;
       } else {
               // log(L_ERR,"MYSQL Connected to %s",mysql_server);
               mysql = &real_mysql_acct;
               break;
       }
     }
       i++;
  }

  if (mysql != NULL) {
     i = 0; 
     while(i<10) { // Try it 10 Times 
      if (type == MYSQL_AUTH) {
        if  (mysql_select_db(mysql, auth_db)) {
               log(L_ERR,"MYSQL cannot select db %s",auth_db);
               log(L_ERR,"MYSQL error: %s",mysql_error(mysql));
        } else {
            //log(L_INFO,"MYSQL Connected to db %s",auth_db);
            return mysql;
        }
      } else {
        if  (mysql_select_db(mysql, acct_db)) {
               log(L_ERR,"MYSQL cannot select db %s",acct_db);
               log(L_ERR,"MYSQL error: %s",mysql_error(mysql));
        } else {
            //log(L_INFO,"MYSQL Connected to db %s",acct_db);
            return mysql;
        }
      } 
       i++;
     }
  }
  log(L_ERR,"MYSQL: Giving up on connect");
  return mysql;
}


/*************************************************************************
 *
 *	Function: my_mysql_query
 *
 *	Purpose: Query MySQL database 
 *
 *************************************************************************/

int my_mysql_query(const char *q, MYSQL *mysql, int type)
{
  int ret;
  int i;
#ifdef STOP_SIGNALS
  sigset_t set,old_set;
#endif

  i = 0; 
  while(i<10) { // Try it 10 Times 
       if (mysql == NULL) mysql = mysql_Reconnect(type);
#ifdef STOP_SIGNALS
       /* Try to not catch any signals during mysql_query 
        * We need this if mysql is less than 3.20.25 or 3.21.3 */
       sigemptyset(&set);
       sigaddset(&set,SIGALRM);                  
       sigprocmask(SIG_BLOCK,&set,&old_set);
#endif
       ret = mysql_query(mysql,q);
#ifdef STOP_SIGNALS
       sigprocmask(SIG_SETMASK,&old_set,NULL);
#endif
       if(!ret) {
         DEBUG("MYSQL query OK");
         return ret;
       }
       if (!strcasecmp(mysql_error(mysql),"mysql server has gone away")) {
               log(L_ERR,"MYSQL Error (retrying %d): Cannot Query:%s",ret,q);
               log(L_ERR,"MYSQL error: %s",mysql_error(mysql));
               mysql_close(mysql);
               mysql = NULL;
       } else {
               log(L_ERR,"MYSQL Error (%d): Cannot Query:%s",ret,q);
               log(L_ERR,"MYSQL error: %s",mysql_error(mysql));
               return ret;
       }
       i++;
  }
  log(L_ERR,"MYSQL Error (giving up): Cannot Query:%s",q);
  return ret;
}

/*************************************************************************
 *
 *	Function: save_mysql_record
 *
 *	Purpose: Saves Account record to MySQL database 
 *
 *************************************************************************/

void save_mysql_record(MYSQL_RECORD * rec)
{

       char            buf[4096];
       char            datebuf[20];
       struct tm      *tim;
       int             err;
       int             num;
       int             time_on;
       int             i = 0;
       time_t          curtime;
       FILE           *sqloutfd;
       MYSQL_RES      *result;
       MYSQL_ROW      row;

       if (use_mysql_acct == 0) return;

       /*
        * Write Detail file.
        */
       sprintf(buf, "%s/sql.log", radacct_dir);
       if ((sqloutfd = fopen(buf, "a")) == (FILE *) NULL) {
               log(L_ERR,"accounting: could not append to file %s", buf);
       }
       if (mysql_acct == NULL)
               mysql_acct = mysql_Reconnect(MYSQL_ACCT);
       
       curtime = time(0) - rec->delay;
       tim = localtime(&curtime);
       strftime(datebuf, sizeof(datebuf), "%Y%m%d%H%M%S", tim);


       if (rec->start == PW_STATUS_ACCOUNTING_ON || rec->start == PW_STATUS_ACCOUNTING_OFF) {
         /* The Terminal server informed us that it was rebooted 
          * -> delete all records */
          if (mysql_acct != NULL) {
               sprintf(buf, "SELECT id,name,unix_timestamp(start_time) FROM %s WHERE server='%s' AND stop_time=0 AND time_on=0", mysql_acct_table, rec->server);

	       DEBUG("%s",buf);
               (void) my_mysql_query(buf, mysql_acct, MYSQL_ACCT);
               if (!(result = mysql_store_result(mysql_acct)) && mysql_num_fields(mysql_acct)) {
                       log(L_ERR,"MYSQL Error: Cannot get result");
                       log(L_ERR,"MYSQL error: %s",mysql_error(mysql_acct));
                       mysql_close(mysql_acct);
                       mysql_acct = NULL;
               } else {
                 num = mysql_num_rows(result);
                 while (i < num) {
                      row = mysql_fetch_row(result);
                      time_on = (int) difftime(curtime,((time_t) atoi(row[2])));
                      sprintf(buf, "UPDATE %s SET stop_time=%s,time_on=%d,term_cause=7 WHERE time_on=0 AND stop_time=0 AND server= '%s' AND id='%s' AND name='%s'", mysql_acct_table,datebuf,time_on,rec->server,row[0],row[1]);

                      (void) my_mysql_query(buf, mysql_acct, MYSQL_ACCT);

                      if (sqloutfd) {
                        fputs(buf, sqloutfd);
                        fputs(";\n", sqloutfd);
                        fflush(sqloutfd);
                      }
                      i++;
                 }
                 mysql_free_result(result);
                }
          return;
        }
        log(L_ERR,"MYSQL Error: Cannot update after %s rebooted",rec->server);
        return;
       }

       if (mysql_acct != NULL) {
               sprintf(buf, "SELECT id FROM %s WHERE id='%s' AND server='%s' AND name='%s'", mysql_acct_table, rec->id, rec->server, rec->name);

	       DEBUG("%s",buf);
               (void) my_mysql_query(buf, mysql_acct, MYSQL_ACCT);
               if (!(result = mysql_store_result(mysql_acct)) && mysql_num_fields(mysql_acct)) {
                       log(L_ERR,"MYSQL Error: Cannot get result");
                       log(L_ERR,"MYSQL error: %s",mysql_error(mysql_acct));
                       mysql_close(mysql_acct);
                       mysql_acct = NULL;
               } else {
                 num = mysql_num_rows(result);
                 mysql_free_result(result);
               }
       }
       sprintf(buf, "INSERT INTO %s (id, time_on, name, server, port, port_type
,
service_type, protocol, ip, call_to, call_from, InBytes, OutBytes, term_cause,
auth) VALUES ('%s',%ld,'%s','%s',%i,%i,%i,%i,'%s','%s','%s',%u,%u,%i,%i)",
               mysql_acct_table,
               rec->id,
               rec->time_used,
               rec->name,
               rec->server,
               rec->port,
               rec->port_type,
               rec->service_type,
               rec->protocol,
               rec->ip,
               rec->call_to,
               rec->call_from,
               rec->inbytes,
               rec->outbytes,
               rec->term_cause,
               rec->auth
               );
	       DEBUG("%s",buf);

       if ((mysql_acct != NULL) && (num == 0)) { //Only insert it if its not there
               if (sqloutfd) {
                  fputs(buf, sqloutfd);
                  fputs(";\n", sqloutfd);
                  fflush(sqloutfd);
               }
			   
               err = my_mysql_query(buf, mysql_acct, MYSQL_ACCT);

               if ((rec->start == 1) && err) {
                       log(L_ERR,"MYSQL Error: Cannot insert");
               }
       }
       if (rec->start == 1) {
               sprintf(buf, "UPDATE %s SET start_time=%s,connect_info='%s' WHERE id='%s' AND server='%s' AND name='%s'",
                       mysql_acct_table,
                       datebuf,
                       rec->connect_info,
                       rec->id, rec->server, rec->name);
       } else {
/* gf */
		if (rec->time_used != 0) {
		    sprintf(buf, "UPDATE %s.%s SET %s.%s.time_used=%s.%s.time_used+%lu WHERE %s.%s.username='%s'",
			auth_db,mysql_auth_table,

			auth_db,mysql_auth_table,

			auth_db,
mysql_auth_table,
			rec->time_used,
			auth_db,mysql_auth_table,
			rec->name );
		    if (mysql_acct != NULL) { 
			my_mysql_query(buf, mysql_acct, MYSQL_ACCT);
		    };
		    DEBUG("%s",buf);
		}
/* gf */
               sprintf(buf, "UPDATE %s SET time_on=%lu,stop_time=%s,inbytes=%u,outbytes=%u,term_cause=%i WHERE id='%s' AND server='%s' AND name='%s'",
                       mysql_acct_table, rec->time_used, datebuf, rec->inbytes,
                       rec->outbytes, rec->term_cause, rec->id, rec->server, 
                       rec->name);
       }
       DEBUG("%s",buf);
       if (sqloutfd) {
               if ( (rec->start != 1) && (num == 0) ) fprintf(sqloutfd,"-- Reco
rd is missing\n");
               fputs(buf, sqloutfd);
               fputs(";\n", sqloutfd);
               fflush(sqloutfd);
       }
       if (mysql_acct != NULL) { 
         if (my_mysql_query(buf, mysql_acct, MYSQL_ACCT)) {
         } else { 
               if (mysql_affected_rows(mysql_acct) !=1) {
                         log(L_ERR,"MYSQL %s (%s) %d rows changed",rec->name,
                             rec->id,mysql_affected_rows(mysql_acct));
               }
         }
       }
       fclose(sqloutfd);

       if(mysql_keepopen == 0) { 
         /* Close the connection after each record */
               if (mysql_acct != NULL) {
                       mysql_close(mysql_acct);
               }
               mysql_acct = NULL;
       }


}


/*************************************************************************
 *
 *	Function: mysql_getpwnam
 *
 *	Purpose: Gets and caches passwd struct from MySQL database 
 *
 *************************************************************************/

struct passwd *mysql_getpwnam (char *name)
{
    static struct passwd *lastpwd;
    static char lastname[64];
    static time_t lasttime = 0;
    time_t now;
    /*char buf[MAX_AUTH_QUERY_LEN];*/
    char *query;
    MYSQL_RES	*result;
    MYSQL_ROW   row;
    now = time(NULL);

    if ((now <= lasttime + 5 ) && strncmp(name, lastname, 64) == 0)
	return lastpwd;

    strncpy(lastname, name, 63);
    lastname[63] = 0;

    if (mysql_auth == NULL) {
        mysql_auth = mysql_Reconnect(MYSQL_AUTH);
    }

    if (mysql_auth != NULL) {
/*
 *  FIXME: This should be more dynamic
 
        sprintf(buf,"SELECT username,passwd,uid,gid FROM %s WHERE username = '%s'", mysql_auth_table, name);

*/
    query = make_query(auth_query, name);

    (void) my_mysql_query(query,mysql_auth,MYSQL_AUTH);
    if (!(result = mysql_store_result(mysql_auth))) {
        log(L_ERR,"MYSQL Error: Could get result for authentication");
        return NULL;
    }
        
    if (mysql_num_rows(result) != 1)
    {
      /* user not found in database */
      return NULL;
    }
    row = mysql_fetch_row(result);

    /*
     * Check for trailing spaces in username since mysql will match a string
     * that has trailing spaces.
     */
    if (strcmp(name,row[0]) != 0) 
    {
     log(L_AUTH,"%s must have trailing space in login attempt",name);
     return NULL;
    }

/*
 *  FIXME:??? Should we fill the entire passwd struct?
 */

    if((lastpwd = malloc(sizeof(struct passwd))) == (struct passwd*)NULL) {
                  log(L_ERR|L_CONS, "no memory");
                exit(1);
        }
    if((lastpwd->pw_name = malloc(sizeof(name)+1)) == (char *)NULL) {
                  log(L_ERR|L_CONS, "no memory");
                exit(1);
        }
    if((lastpwd->pw_passwd = malloc(sizeof(char[15]))) == (char *)NULL) {
                  log(L_ERR|L_CONS, "no memory");
                exit(1);
        }

    strcpy (lastpwd->pw_name , name);
    strcpy (lastpwd->pw_passwd , row[1]);
    lastpwd->pw_uid = (uid_t) atoi(row[2]);
    lastpwd->pw_gid = (gid_t) atoi(row[3]);
    lastpwd->pw_gecos = NULL;
    lastpwd->pw_dir = NULL;
    lastpwd->pw_shell = NULL;


    lasttime = now;
     
    mysql_free_result(result);

    if(mysql_keepopen == 0) {
         /* Close the connection after each auth query */
               if (mysql_auth != NULL) {
                       mysql_close(mysql_auth);
               }
               mysql_auth = NULL;
       }


    return lastpwd;
   } else
    return NULL;
}

/*************************************************************************
 *
 *	Function: mysql_pass
 *
 *	Purpose: Check the user's password against MySQL database 
 *
 *************************************************************************/

int mysql_pass(char *name, char *passwd)
{
    struct passwd *pwd;
    char *encr_pass;
    if (use_mysql_auth == 0) 
    {
      log(L_ERR,"MySQL Auth specified in users file, but not in sqlserver file");
      return -1;
    }

    if ((pwd = mysql_getpwnam(name)) == NULL)
     {
       return -1;
     }

    encr_pass = crypt(passwd,pwd->pw_passwd);


    if (strcmp(pwd->pw_passwd,encr_pass))
      {
       /* password mismatch */
       return -1;
      }

    DEBUG("User %s successfully authenticated via mysql", name);
    return 0;


}


/*
 *	Replace %<whatever> in a string.
 *
 *	%u   User name
 *
 */
static char *make_query(char *str, char *name)
{
	static char buf[MAX_AUTH_QUERY_LEN];
	int i = 0, c;
	char *p;

	for(p = str; *p; p++) {
		c = *p;
		if (c != '%' && c != '\\') {
			buf[i++] = *p;
			continue;
		}
		if (*++p == 0) break;
		if (c == '%') switch(*p) {
			case '%':
				buf[i++] = *p;
				break;
			case 'u': /* User name */
				if (name != NULL)
					strcpy(buf + i, name);
				else
					strcpy(buf + i, " ");
				i += strlen(buf + i);
				break;
			default:
				buf[i++] = '%';
				buf[i++] = *p;
				break;
		}
		if (c == '\\') switch(*p) {
			case 'n':
				buf[i++] = '\n';
				break;
			case 'r':
				buf[i++] = '\r';
				break;
			case 't':
				buf[i++] = '\t';
				break;
			default:
				buf[i++] = '\\';
				buf[i++] = *p;
				break;
		}
	}
	if (i >= MAX_AUTH_QUERY_LEN)
		i = MAX_AUTH_QUERY_LEN - 1;
	buf[i++] = 0;

	return buf;
}

static  void fieldcpy(char *string, char **uptr)
{
        char    *ptr;

        ptr = *uptr;
        while (*ptr == ' ' || *ptr == '\t') {
              ptr++;
        }
        if(*ptr == '"') {
                ptr++;
                while(*ptr != '"' && *ptr != '\0' && *ptr != '\n') {
                        *string++ = *ptr++;
                }
                *string = '\0';
                if(*ptr == '"') {
                        ptr++;
                }
                *uptr = ptr;
                return;
        }

        while(*ptr != ' ' && *ptr != '\t' && *ptr != '\0' && *ptr != '\n' &&
                                                *ptr != '=' && *ptr != ',') {
                        *string++ = *ptr++;
        }
        *string = '\0';
        *uptr = ptr;
	return;
}

#endif /* USEMYSQL */

